//******************************************************************************
//  TVideoCapture -  wrapper for DirectShow video capture functions
//  Based on
//  Microsoft's AMCap & StillCap
//  and
//  DScapture by orthkon * www.mp3.com/orthkon * orthkon@mail.com
//
//  written by Egor Averchenkov,  e_g_o_r@mail.ru
//  version 1.07
//
//  history log:
//  01-04-24 - v1.00 (initial release)
//
//  01-09-18 - v1.02 -  I had to continue this work after a long delay
//    some stupid bugs corrected
//    interface section was slightly ordered
//    new property CapturePixelFormat added to select color depth of captured bitmaps
//
//  01-09-20 - v1.03 - orthkon mailed me TCapture based on this component - good work!
//    added properties and functions to work with available capture modes
//
//  01-09-22 - v1.04
//    working with filters property pages
//
//  01-10-18 - v1.05
//    working with compressors
//
//  01-10-27 - v1.06 - thanks to Lee_Nover (Lee.Nover@email.si) - he did all dirty work))))
//    one graph for preview/capture
//    compression dialogs
//
//  01-10-28 - v1.07
//    bugs in building graph with compressors were corrected
//
//******************************************************************************
unit VCap;

interface

uses
  Windows, Messages, SysUtils, Classes, Graphics, Controls, Forms, Dialogs,
  extctrls, DirectShow, ActiveX, DirectSound;

const
  WM_FGNOTIFY = WM_USER + 1;
  WM_SAMPLECAPTURED = WM_USER + 2;

type
  EVideoCaptureError = class(Exception);

type
  TBitmapCapturedEvent = procedure(CapturedImage: TBitmap) of object;

type
  TDVSize = (dvsDontWorry, dvsFull, dvsHalf, dvsQuater, dvsDC);

const
  DVSizeName: array [TDVSize] of string =
  ('Doesn''t matter', 'Full size', 'Half size', 'Quater', '1/8');
{
const
  DVSizeName: array [TDVSize] of string =
  ('  ', ' ', '', '', '1/8 ');
}

// -----------------------------------------------------------------------------
// Working with capture modes

type
  TVCapMode = record
    MediaType:    TGUID;
    MediaSubType: TGUID;
    Width:        integer;
    Height:       integer;
    BitCount:     integer;
    FrameRate,
    MinFrameRate,
    MaxFrameRate: double;
  end;

type
  TACapMode = record
    MediaType:    TGUID;
    MediaSubType: TGUID;
    SampleRate:   integer;
    SampleSize:   integer;
    Channels:     integer;
  end;

function IsEqualModes(const mode1, mode2: TVCapMode): boolean; overload;
function IsEqualModes(const mode1, mode2: TACapMode): boolean; overload;

function GetModeString(const mode: TVCapMode): string; overload;
function GetModeString(const mode: TACapMode): string; overload;

// -----------------------------------------------------------------------------
// Various Capture Dialogs
type
  TCaptureDialog = (
    cdVFORMAT,
    cdVSOURCE,
    cdVDISPLAY,
    cdVCAPTURE,
    cdVCROSSBAR,
    cdTVTUNER,
    cdACAPTURE,
    cdACROSSBAR,
    cdTVAUDIO,
    cdVCAPTURE_PIN,
    cdVPREVIEW_PIN,
    cdACAPTURE_PIN,
    cdVCOMPRESSION,
    cdACOMPRESSION
  );
  TCaptureDialogs = set of TCaptureDialog;

const
  DialogTitles: array [TCaptureDialog] of string = (
    'Video Format',	        // VFORMAT,
    'Video Source',	        // VSOURCE,
    'Video Display',	        // VDISPLAY,
    'Video Capture Filter',	// VCAPTURE,
    'Video Crossbar',           // VCROSSBAR,
    'TV Tuner',	                // TVTUNER,
    'Audio Capture Filter',	// ACAPTURE,
    'Audio Crossbar',           // ACROSSBAR,
    'TV Audio',	                // TVAUDIO,
    'Video Capture Pin',	// VCAPTURE_PIN,
    'Video Preview Pin',	// VPREVIEW_PIN,
    'Audio Capture Pin',        // ACAPTURE_PIN
    'Video Compressor',	        // VCOMPRESSOR,
    'Audio Compressor'	        // ACOMPRESSOR,
  );

// -----------------------------------------------------------------------------
// TVideoCapture

type
  TVideoCapture = class(TCustomControl)
  private
    Graph: IGraphBuilder;
    Builder: ICaptureGraphBuilder2;
    VideoWindow: IVideoWindow;
    MediaEvent: IMediaEventEx;
    DroppedFrames: IAMDroppedFrames;
    VideoCompression: IAMVideoCompression;
    CaptureDialogs: IAMVfwCaptureDialogs;
    AStreamConf: IAMStreamConfig;      // for audio cap
    VStreamConf: IAMStreamConfig;      // for video cap
    Render: IBaseFilter;
    VCap,  ACap,
    VComp, AComp: IBaseFilter;
    Sink: IFileSinkFilter;
    ConfigAviMux: IConfigAviMux;
    Grabber: ISampleGrabber;

    GrabberCB: TObject;

    fCapAudioIsRelevant: boolean;
    FCapAudio: boolean;
    fCCAvail: boolean;
    fCapCC: boolean;
    fGraphBuilt: boolean;
    fPreviewFaked: boolean;
    FACapFriendlyName,
    FVCapFriendlyName: string;
    FCaptureWidth,
    FCaptureHeight: integer;

    FCapturing: boolean;
    FPreviewing: boolean;
    FUseFrameRate: boolean;
    FUseTimeLimit: boolean;
    FWantPreview: boolean;
    FFrameRate: double;
    FCapStartTime: DWORD;
    FCapStopTime: DWORD;
    FMasterStream: integer;

    FNotDropped: integer;
    FDroppedFrames: integer;
    FNotDroppedBase: integer;
    FDroppedBase: integer;
    FCapTime: DWORD;

    FTempCaptureFileName: string;
    FCaptureFileName: string;
    mVideo, mAudio,
    mVComp, mAComp: IMoniker;
    FCaptureTimer: TTimer;

    FTimeLimit: integer;
    FUseTempFile: boolean;
    FPreallocFileSize: Cardinal;
    FCaptureFile: WideString;
    FWantBitmaps: boolean;
    FDVSize: TDVSize;

    FOnStopPreview: TNotifyEvent;
    FOnStartPreview: TNotifyEvent;
    FOnStopCapture: TNotifyEvent;
    FOnStartCapture: TNotifyEvent;
    FOnChangeDevice: TNotifyEvent;
    FOnCaptureProgress: TNotifyEvent;
    FOnBitmapGrabbed: TBitmapCapturedEvent;
    FPixelFormat: TPixelFormat;
    FOnChangeACapMode: TNotifyEvent;
    FOnChangeVCapMode: TNotifyEvent;
    FDialogs: TCaptureDialogs;
    FOnChangeCompressor: TNotifyEvent;
    FDoPreallocFile: boolean;
    FWantCapture: boolean;
    FWantAudio: boolean;

    procedure SetMasterStream(const Value: integer);
    property MasterStream: integer read FMasterStream write SetMasterStream;
    procedure SetFrameRate(const Value: double);
    procedure SetTempCaptureFileName(const Value: string);
    procedure SetCaptureFileName(const Value: string);
    function GetACapMode: TACapMode;
    function GetACapModeCount: integer;
    function GetACapModes(i: integer): TACapMode;
    function GetVCapMode: TVCapMode;
    function GetVCapModeCount: integer;
    function GetVCapModes(i: integer): TVCapMode;
  private
    procedure SetSize(var msg: TMessage); message WM_SIZE;  // Changing size of cap window
    procedure GraphEvent(var msg: TMessage); message WM_FGNOTIFY;
    procedure BitmapGrabbed(var msg: TMessage); message WM_SAMPLECAPTURED;
    procedure CaptureProgress(Sender: TObject);
    procedure ResizeWindow;
    procedure UpdateStatus;
    procedure ChooseDevices(nmVideo, nmAudio: IMoniker; ForceRebuild: boolean = false); overload;
    procedure ChooseCompressors(nmVComp, nmAComp: IMoniker; ForceRebuild: boolean = false); overload;

    procedure CleanUp;
    procedure FreeCapFilters;
    function InitCapFilters: boolean;
    function MakeBuilder: boolean;
    function MakeGraph: boolean;
    procedure NukeDownstream(pf: IBaseFilter);
    procedure TearDownGraph;
    function RenderPreviewPin: boolean;
    function FindVideoWindow: boolean;
    function AdjustGrabberPosition: boolean;
    function BuildGraph: boolean;
    function AllocCaptureFile(const SizeMb: integer): boolean;
    function SaveCaptureFile(const FileName: string): boolean;
    procedure GetAvailableDialogs;
  public
    // if someone wants to work with filters
    property VCapFilter: IBaseFilter read VCap;
    property ACapFilter: IBaseFilter read ACap;
    property VCompFilter: IBaseFilter read VComp;
    property ACompFilter: IBaseFilter read AComp;

    // state properties
    property Capturing: boolean read FCapturing;
    property Previewing: boolean read FPreviewing;
    // video and audio capture properties
    property VCapFriendlyName: string read FVCapFriendlyName;
    property VCapMode: TVCapMode read GetVCapMode;
    property VCapModeCount: integer read GetVCapModeCount;
    property VCapModes[i: integer]: TVCapMode read GetVCapModes;
    function SetVCapMode(i: integer): boolean; overload;
    function SetVCapMode(const mode: TVCapMode): boolean; overload;

    property ACapFriendlyName: string read FACapFriendlyName;
    property ACapMode: TACapMode read GetACapMode;
    property ACapModeCount: integer read GetACapModeCount;
    property ACapModes[i: integer]: TACapMode read GetACapModes;
    function SetACapMode(i: integer): boolean; overload;
    function SetACapMode(const mode: TACapMode): boolean; overload;

    property Dialogs: TCaptureDialogs read FDialogs;
    property FramesDropped: integer read FDroppedFrames;
    property NotDropped: integer read FNotDropped;
    property CapStartTime: DWORD read FCapStartTime;
    property CapStopTime: DWORD read FCapStopTime;
    property CapTime: DWORD read FCapTime;

    procedure ChooseDevices(const szVideo, szAudio: string; ForceRebuild: boolean = false); overload;
    procedure ChooseDevices(const numVideo, numAudio: integer; ForceRebuild: boolean = false); overload;

    procedure ChooseCompressors(const szVComp, szAComp: string; ForceRebuild: boolean = false); overload;
    procedure ChooseCompressors(const numVComp, numAComp: integer; ForceRebuild: boolean = false); overload;

    function Init: boolean;
    function StartPreview: boolean;
    function StopPreview: boolean;
    function StartCapture(const Dialog: boolean = false): boolean;
    function StopCapture: boolean;
    function CaptureFrame: boolean;

    function ShowDialog(const DialogType: TCaptureDialog): boolean;

    constructor Create(AOwner: TComponent); override;
    destructor Destroy; override;
  published
    property DVPreviewSize: TDVSize read FDVSize write FDVSize;

    // set to true if you want to capture single frames during preview
    property WantBitmaps: boolean read FWantBitmaps write FWantBitmaps;
    property CapturePixelFormat: TPixelFormat read FPixelFormat write FPixelFormat;

    property WantAudio: boolean read FWantAudio write FWantAudio; // set true if you want audio in captured file
    property WantPreview: boolean read FWantPreview write FWantPreview;
    property WantCapture: boolean read FWantCapture write FWantCapture; // set to false if dont want to capture file is being created
    property UseFrameRate: boolean read FUseFrameRate write FUseFrameRate;
    property FrameRate: double read FFrameRate write SetFrameRate;
    property CaptureFileName: string read FCaptureFileName write SetCaptureFileName;
    property UseTempFile: boolean read FUseTempFile write FUseTempFile;
    property TempCaptureFileName: string read FTempCaptureFileName write SetTempCaptureFileName;
    property UseTimeLimit: boolean read FUseTimeLimit write FUseTimeLimit;
    property TimeLimit: integer read FTimeLimit write FTimeLimit;
    property DoPreallocFile: boolean read FDoPreallocFile write FDoPreallocFile;
    property PreallocFileSize: Cardinal read FPreallocFileSize write FPreallocFileSize default 10;

    property OnChangeDevice: TNotifyEvent read FOnChangeDevice write FOnChangeDevice;
    property OnChangeCompressor: TNotifyEvent read FOnChangeCompressor write FOnChangeCompressor;
    property OnChangeVCapMode: TNotifyEvent read FOnChangeVCapMode write FOnChangeVCapMode;
    property OnChangeACapMode: TNotifyEvent read FOnChangeACapMode write FOnChangeACapMode;
    property OnCaptureProgress: TNotifyEvent read FOnCaptureProgress write FOnCaptureProgress;
    property OnStopCapture: TNotifyEvent read FOnStopCapture write FOnStopCapture;
    property OnStopPreview: TNotifyEvent read FOnStopPreview write FOnStopPreview;
    property OnStartCapture: TNotifyEvent read FOnStartCapture write FOnStartCapture;
    property OnStartPreview: TNotifyEvent read FOnStartPreview write FOnStartPreview;
    property OnBitmapGrabbed: TBitmapCapturedEvent read FOnBitmapGrabbed write FOnBitmapGrabbed;
  published
    property Align;
    property Color;
    property Visible;
    property OnMouseMove;
    property OnMouseUp;
    property OnMouseDown;
    property OnClick;
    property OnDblClick;
 end;

// device enum functions
// caller have to free aquired list!!!
function GetVideoDevicesList(const Refresh: boolean = false): TStringList;
function GetAudioDevicesList(const Refresh: boolean = false): TStringList;
function GetVideoCompressorsList(const Refresh: boolean = false): TStringList;
function GetAudioCompressorsList(const Refresh: boolean = false): TStringList;

procedure Register;

implementation

uses MMSystem, contnrs;

//-----------------------------------------------------------------------
procedure Register;
begin
  RegisterComponents('Samples', [TVideoCapture]);
end;

//-----------------------------------------------------------------
resourcestring
{
  rsDShowCapture = 'DirectShow - ';
  rsGraphCantPreview = '     !';
  rsCantRenderCC = '   closed captioning!';
  rsGraphCantBePreviewedProperly = '     !';
  rsCantSetPreviewFrameRate = '%x:       !';
  rsSetMasterStreamFailed = ' SetMasterStream!';
  rsCantSetCaptureFile = '    !';
  rsCantRenderVCaptureStream = '    !';
  rsCantRenderPreviewStream = '    !';
  rsCantRenderACaptureStream = '    !';
  rsThisGraphCantPreview = '     !';
  rsCantSetCaptureFrameRate = '      !';
  rsCantMakeGraphBuilder = '    .    DirectShow!';
  rsCantCreateVCaptureFilter = ' %x.'#13#10'     -    !';
  rsCantMakeGraph = '   .    DirectShow!';
  rsCantSetFilterGraph = '    !';
  rsCantAddVFilterToGraph = ' %x:       !';
  rsCantFindVStreamConfig = ' %x:    VCapture:IAMStreamConfig!';
  rsCantMakeACapFilter = '    !';
  rsCantAddAFilterToGraph = ' %x:       !';
  rsCantFindAStreamConfig = '   ACapture:IAMStreamConfig';
  rsCantRunPreviewGraph = ' %x:     !';
  rsCantStopPreviewGraph = ' %x:     !';
  rsCantGetMediaControl = ' %x:    IMediaControl';
  rsCantRunGraph = ' %x:    !';
  rsStartCapture = ' !';
  rsCantStopGraph = ' %x:    !';
  rsEmptyFileName = '      !';
  rsFailedToAllocFileSize = '     !  ?';
//  rsDShow80Required = '     DirectShow 8.0!';
}
  rsDShowCapture = 'DirectShow - Capture';
  rsCantRenderCC = 'Cannot render closed captioning!';
  rsCantSetPreviewFrameRate = '%x: can''t set preview frame rate!';
  rsSetMasterStreamFailed = 'SetMasterStream failed!';
  rsCantSetCaptureFile = 'Can''t set capture file!';
  rsCantRenderVCaptureStream = 'Can''t render video capture stream!';
  rsCantRenderPreviewStream = 'Can''t render preview stream!';
  rsCantRenderACaptureStream = 'Can''t render audio preview stream!';
  rsThisGraphCantPreview = 'This graph cannot preview!';
  rsCantSetCaptureFrameRate = 'Cannot set frame rate for capture!';
  rsCantMakeGraphBuilder = 'Can''t init graph builder. Probably DirectShow is not installed!';
  rsCantCreateVCaptureFilter = 'Error %x.'#13#10'Can''t create video capture filter - propably you havn''t video capture device!';
  rsCantMakeGraph = 'Can''t init graph. Probably DirectShow is not installed!';
  rsCantSetFilterGraph = 'Can''t set filter graph!';
  rsCantAddVFilterToGraph = 'Error %x: Can''t add video capture filter into graph!';
  rsCantFindVStreamConfig = 'Error %x: Can''t find VCapture:IAMStreamConfig!';
  rsCantMakeACapFilter = 'Can''t create audio capture filter!';
  rsCantAddAFilterToGraph = 'Error %x: Can''t add audio capture filter into graph!';
  rsCantFindAStreamConfig = 'Can''t find ACapture:IAMStreamConfig!';
  rsCantRunPreviewGraph = 'Error %x: Cannot run preview graph!';
  rsCantStopPreviewGraph = 'Error %x: Cannot stop preview graph!';
  rsCantGetMediaControl = 'Error %x: Can''t get IMediaControl!';
  rsCantRunGraph = 'Error %x: Cannot run graph!';
  rsStartCapture = 'Starting capture!';
  rsCantStopGraph = 'Error %x: Cannot stop graph!';
  rsEmptyFileName = 'Capture filename required!';
  rsFailedToAllocFileSize = 'Can''t allocate space for capture file! Disk full?';
  rsCantInstallSampleGrabber = 'Can''t install SampleGrabber. Probably wrong color depth using?';
  rsCantAddVCompFilter = 'Can''t add video compression';
  rsCantAddACompFilter = 'Can''t add audio compression';

//-----------------------------------------------------------------
const
  MAX_TIME = $7FFFFFFFFFFFFFFF;

const
  VCompFilterName = 'VideoCompressor';
  ACompFilterName = 'AudioCompressor';

const
  IID_IPropertyBag: TGUID = '{55272A00-42CB-11CE-8135-00AA004BB851}';
  IID_ISpecifyPropertyPages: TGUID = '{B196B28B-BAB4-101A-B69C-00AA00341D07}';

type
  PVIDEOINFOHEADER = ^TVIDEOINFOHEADER;
  TVIDEOINFOHEADER = record
    rcSource: TRECT;          // The bit we really want to use
    rcTarget: TRECT;          // Where the video should go
    dwBitRate: DWORD;         // Approximate bit data rate
    dwBitErrorRate: DWORD;    // Bit error rate for this stream
    AvgTimePerFrame: TReference_Time;   // Average time per frame (100ns units)
    bmiHeader: BITMAPINFOHEADER;
  end;

  PVIDEOINFOHEADER2 = ^TVIDEOINFOHEADER2;
  TVIDEOINFOHEADER2 = record
    rcSource: TRECT;          // The bit we really want to use
    rcTarget: TRECT;          // Where the video should go
    dwBitRate: DWORD;         // Approximate bit data rate
    dwBitErrorRate: DWORD;    // Bit error rate for this stream
    AvgTimePerFrame: TReference_Time;   // Average time per frame (100ns units)
    dwInterlaceFlags: DWORD;
    dwCopyProtectFlags: DWORD;
    dwPictAspectRatioX: DWORD;
    dwPictAspectRatioY: DWORD;
    dwReserved1: DWORD;
    dwReserved2: DWORD;
    bmiHeader: BITMAPINFOHEADER;
  end;

  PWAVEFORMATEX = ^TWAVEFORMATEX;
  TWAVEFORMATEX = record
    wFormatTag: WORD;
    nChannels: WORD;
    nSamplesPerSec: DWORD;
    nAvgBytesPerSec: DWORD;
    nBlockAlign: WORD;
    wBitsPerSample: WORD;
    cbSize: WORD;
  end;

const
  DVSizes: array [TDVSize] of integer = (
    0,
    DVRESOLUTION_FULL,
    DVRESOLUTION_HALF,
    DVRESOLUTION_QUARTER,
    DVRESOLUTION_DC
  );

//-----------------------------------------------------------------
var
  VideoDevicesList,
  AudioDevicesList,
  VideoCompressorsList,
  AudioCompressorsList: TObjectList;

//-----------------------------------------------------------------
function IsEqualModes(const mode1, mode2: TVCapMode): boolean;
begin
  Result:= (mode1.Height = mode2.Height) and
           (mode1.Width = mode2.Width) and
           (mode1.BitCount = mode2.BitCount);
end;

function IsEqualModes(const mode1, mode2: TACapMode): boolean;
begin
  Result:= (mode1.SampleRate = mode2.SampleRate) and
           (mode1.SampleSize = mode2.SampleSize) and
           (mode1.Channels = mode2.Channels);
end;

function GetModeString(const mode: TVCapMode): string; overload;
begin
  Result:= IntToStr(mode.Width)+'x'+IntToStr(mode.Height)+'x'+IntToStr(mode.BitCount)+'b';
end;

function GetModeString(const mode: TACapMode): string; overload;
begin
  Result:= IntToStr(mode.SampleRate)+'Hz x '+IntToStr(mode.SampleSize)+'bit x '+IntToStr(mode.Channels);
end;

//-----------------------------------------------------------------
function MyMsg(szMsg: string; hr: HRESULT): string;
begin
  Result:= Format(szMsg, [hr]);
  MessageBox(GetForegroundWindow, PChar(Result), PChar(rsDShowCapture), MB_OK or MB_ICONSTOP);
end;

procedure ErrMsg(szMsg: string; hr: HRESULT = 0);
begin
  MyMsg(szMsg, hr);
end;

procedure ErrMsgException(szMsg: string; hr: HRESULT = 0);
begin
  raise EVideoCaptureError.Create(MyMsg(szMsg, hr));
end;

//-----------------------------------------------------------------

function CheckGUID(p1, p2: TGUID): boolean;
var
  i: byte;
begin
  Result:= false;
  for i:= 0 to 7 do if p1.D4[i] <> p2.D4[i] then Exit;
  Result:= (p1.D1 = p2.D1) and (p1.D2 = p2.D2) and (p1.D3 = p2.D3);
end;

// Free an existing media type (ie free resources it holds)
procedure FreeMediaType(mt: TAM_MEDIA_TYPE);
begin
  if mt.cbFormat <> 0 then begin
    CoTaskMemFree(mt.pbFormat);
    // Strictly unnecessary but tidier
    mt.cbFormat:= 0;
    mt.pbFormat:= nil;
  end;
  mt.pUnk:= nil;
end;

procedure DeleteMediaType(pmt: PAM_MEDIA_TYPE);
begin
  // allow NULL pointers for coding simplicity
  if pmt <> nil then begin
    FreeMediaType(pmt^);
    CoTaskMemFree(pmt);
  end;
end;

//-----------------------------------------------------------------
{ TCapDeviceInfo }

type
  TCapDeviceInfo = class
  public
    DeviceName: string;
    DeviceNameInList: string;
    Moniker: IMoniker;
    constructor Create(const aDeviceName, aDeviceNameInList: string; const aMoniker: IMoniker);
    destructor Destroy; override;
  end;

constructor TCapDeviceInfo.Create(const aDeviceName, aDeviceNameInList: string; const aMoniker: IMoniker);
begin
  DeviceName:= aDeviceName;
  DeviceNameInList:= aDeviceNameInList;
  Moniker:= aMoniker;
end;

destructor TCapDeviceInfo.Destroy;
begin
  Moniker:= nil; // I guess we have to release moniker - tnx to Gabriel Corneanu
  inherited;
end;

//-----------------------------------------------------------------

function FindDeviceNum(const DeviceName: string; Devices: TObjectList): integer;
var
  i: integer;
  d: TObject;
begin
  Result:= 1;
  for i:= 0 to Devices.Count-1 do begin
    d:= Devices[i];
    if (d is TCapDeviceInfo) and (TCapDeviceInfo(d).DeviceName=DeviceName) then begin
      Result:= Result + 1;
    end;
  end;
end;

procedure EnumFilters(const clsidDeviceClass: TGUID; const FiltersList: TObjectList);
var
  SysDevEnum: ICreateDevEnum;
  EnumCat: IEnumMoniker;
  Moniker: IMoniker;
  cFetched: Longint;
  PropBag: IPropertyBag;
  varName: OleVariant;
  hr: HRESULT;
  n: integer;
  s: string;
begin
  FiltersList.Clear;
  SysDevEnum:= nil;
  hr:= CoCreateInstance(CLSID_SystemDeviceEnum, nil, CLSCTX_INPROC, IID_ICreateDevEnum, SysDevEnum);
  if (hr = S_OK) then begin
    // enum available video capture devices
    EnumCat:= nil;
    if (SysDevEnum.CreateClassEnumerator(clsidDeviceClass, EnumCat, 0)= S_OK) then
      while EnumCat.Next(1, Moniker, @cFetched) = S_OK do begin
        Moniker.BindToStorage(nil, nil, IID_IPropertyBag, PropBag);
        PropBag.Read('FriendlyName', varName, nil);
        s:= varName;
        n:= FindDeviceNum(s, FiltersList);
        if n>1 then s:= s + ' ('+IntToStr(n)+')';
        FiltersList.Add(TCapDeviceInfo.Create(varName, s, Moniker) );
        PropBag:= nil;
        Moniker:= nil;
      end;
    EnumCat:= nil;
  end;
end;

//-----------------------------------------------------------------
// Capture devices Enum functions
procedure BuildCaptureDevicesList;
begin
  EnumFilters(CLSID_VideoInputDeviceCategory, VideoDevicesList);
  EnumFilters(CLSID_AudioInputDeviceCategory, AudioDevicesList);
end;

function GetVideoDevicesList(const Refresh: boolean): TStringList;
var i: integer;
begin
  Result:= TStringList.Create;
  if Refresh or (VideoDevicesList.Count=0) then
    EnumFilters(CLSID_VideoInputDeviceCategory, VideoDevicesList);
  for i:= 0 to VideoDevicesList.Count-1 do
    Result.Add(TCapDeviceInfo(VideoDevicesList[i]).DeviceNameInList);
end;

function GetAudioDevicesList(const Refresh: boolean): TStringList;
var i: integer;
begin
  Result:= TStringList.Create;
  if Refresh or (AudioDevicesList.Count=0) then
    EnumFilters(CLSID_AudioInputDeviceCategory, AudioDevicesList);
  for i:= 0 to AudioDevicesList.Count-1 do
    Result.Add(TCapDeviceInfo(AudioDevicesList[i]).DeviceNameInList);
end;

//-----------------------------------------------------------------
// Compressors Enum functions
function GetVideoCompressorsList(const Refresh: boolean = false): TStringList;
var i: integer;
begin
  Result:= TStringList.Create;
  if Refresh or (VideoCompressorsList.Count=0) then
    EnumFilters(CLSID_VideoCompressorCategory, VideoCompressorsList);
  for i:= 0 to VideoCompressorsList.Count-1 do
    Result.Add(TCapDeviceInfo(VideoCompressorsList[i]).DeviceNameInList);
end;

function GetAudioCompressorsList(const Refresh: boolean = false): TStringList;
var i: integer;
begin
  Result:= TStringList.Create;
  if Refresh or (AudioCompressorsList.Count=0) then
    EnumFilters(CLSID_AudioCompressorCategory, AudioCompressorsList);
  for i:= 0 to AudioCompressorsList.Count-1 do
    Result.Add(TCapDeviceInfo(AudioCompressorsList[i]).DeviceNameInList);
end;

//-----------------------------------------------------------------
type
  TSampleGrabberCB = class(TObject, ISampleGrabberCB)
  private
    FBuffer: PByte;
    FBuffLen: Cardinal;
    FOwner: HWND;
    FEnable: boolean;
  public
    { IUnknown }
    function QueryInterface(const IID: TGUID; out Obj): HRESULT; stdcall;
    function _AddRef: Integer; stdcall;
    function _Release: Integer; stdcall;
    { ISampleGrabberCB }
    function SampleCB(SampleTime: Double; pSample: IMediaSample): HResult; stdcall;
    function BufferCB(SampleTime: Double; pBuffer: PByte; BufferLen: longint): HResult; stdcall;
  public
    property Owner: HWND read FOwner write FOwner;
    property Buffer: PByte read FBuffer;
    property BufferLength: Cardinal read FBuffLen;
    property GrabbingEnabled: boolean read FEnable write FEnable;

    procedure ClearBuffer;

    constructor Create;
    destructor Destroy; override;
  end;

//-----------------------------------------------------------------
{ TSampleGrabberCB }
constructor TSampleGrabberCB.Create;
begin
  FBuffer:= nil;
  FBuffLen:= 0;
end;

destructor TSampleGrabberCB.Destroy;
begin
  ClearBuffer;
  inherited;
end;

procedure TSampleGrabberCB.ClearBuffer;
begin
  FBuffLen:= 0;
  ReallocMem(FBuffer, FBuffLen);
end;

{ TSampleGrabberCB - IUnknown }
function TSampleGrabberCB._AddRef: Integer;
begin
  Result:= 2;
end;

function TSampleGrabberCB._Release: Integer;
begin
  Result:= 1;
end;

function TSampleGrabberCB.QueryInterface(const IID: TGUID; out Obj): HRESULT;
begin
  // We need to return the two event interfaces when they're asked for
  Result:= E_NOINTERFACE;
  if CheckGUID(IID, ISampleGrabberCB) or CheckGUID(IID, IUnknown) then begin
    if GetInterface(IID,Obj) then Result:= S_OK;
  end;
end;

{ TSampleGrabberCB - ISampleGrabberCB }
function TSampleGrabberCB.BufferCB(SampleTime: Double; pBuffer: PByte; BufferLen: Integer): HResult;
begin
  if GrabbingEnabled then begin
// Comment by microsoft's programmer below
    // The sample grabber is calling us back on its deliver thread.
    // This is NOT the main app thread!
    //
    //           !!!!! WARNING WARNING WARNING !!!!!
    //
    // On Windows 9x systems, you are not allowed to call most of the
    // Windows API functions in this callback.  Why not?  Because the
    // video renderer might hold the global Win16 lock so that the video
    // surface can be locked while you copy its data.  This is not an
    // issue on Windows 2000, but is a limitation on Win95,98,98SE, and ME.
    // Calling a 16-bit legacy function could lock the system, because
    // it would wait forever for the Win16 lock, which would be forever
    // held by the video renderer.
    // As a workaround, we will copy the bitmap data during the callback,
    // post a message to our app, and write the data later.
    ReallocMem(FBuffer, BufferLen);
    FBuffLen:= BufferLen;
    CopyMemory(FBuffer, pBuffer, BufferLen);
    PostMessage(FOwner, WM_SAMPLECAPTURED, integer(Self), 0);
    GrabbingEnabled:= false;
  end;
  Result:= S_OK;
end;

function TSampleGrabberCB.SampleCB(SampleTime: Double; pSample: IMediaSample): HResult;
begin // not implemented
  Result:= S_OK;
end;

//-----------------------------------------------------------------
{ TVideoCapture }

constructor TVideoCapture.Create(AOwner: TComponent);
begin
  inherited;
  CleanUp;
  FUseFrameRate:= false;
  FWantAudio:= true;
  FWantPreview:= true;
  FWantCapture:= true;
  Color:= clBlue;
  FWantBitmaps:= true;
  Width:= 100;
  Height:= 100;
  FPreallocFileSize:= 1;
  FPixelFormat:= pf24bit;

  GrabberCB:= TSampleGrabberCB.Create;

  FCaptureTimer:= TTimer.Create(Self);
  FCaptureTimer.Interval:= 100;
  FCaptureTimer.OnTimer:= CaptureProgress;
end;

destructor TVideoCapture.Destroy;
begin
  StopPreview;
  StopCapture;
  TearDownGraph;
  CleanUp;

  FCaptureTimer.Enabled:= false;
  FCaptureTimer.Free;
  GrabberCB.Free;
  inherited;
end;

//-----------------------------------------------------------------
procedure TVideoCapture.ChooseDevices(nmVideo, nmAudio: IMoniker; ForceRebuild: boolean = false);
begin
  if (ForceRebuild) or((mVideo<>nmVideo) or (mAudio<>nmAudio)) then  begin
    mVideo:= nmVideo; mAudio:= nmAudio;
    TearDownGraph;

    FreeCapFilters;
    asm
      //this was added due to corrupted registers after func call (D5 compiler error?)
      push ebx; push edi; push esi
    end;
    InitCapFilters;
    asm
      pop esi; pop edi; pop ebx
    end;

    BuildGraph;
    if FWantPreview then StartPreview;

    if Assigned(FOnChangeDevice) then FOnChangeDevice(Self);
  end;
end;

procedure TVideoCapture.ChooseDevices(const szVideo, szAudio: string; ForceRebuild: boolean = false);
var
  nmVideo, nmAudio: IMoniker;
  i: integer;
begin
  nmVideo:= nil; nmAudio:= nil;

  for i:= 0 to VideoDevicesList.Count-1 do
    with VideoDevicesList[i] as TCapDeviceInfo do
      if DeviceNameInList = szVideo then begin
        nmVideo:= Moniker;
        break;
      end;

  for i:= 0 to AudioDevicesList.Count-1 do
    with AudioDevicesList[i] as TCapDeviceInfo do
      if DeviceNameInList = szAudio then begin
        nmAudio:= Moniker;
        break;
      end;

  ChooseDevices(nmVideo, nmAudio, ForceRebuild);
  nmVideo:= nil; nmAudio:= nil;
end;

procedure TVideoCapture.ChooseDevices(const numVideo, numAudio: integer; ForceRebuild: boolean = false);
var
  nmVideo, nmAudio: IMoniker;
begin
  nmVideo:= nil; nmAudio:= nil;

  if (numVideo>=0) and(numVideo<VideoDevicesList.Count) then
    nmVideo:= TCapDeviceInfo(VideoDevicesList[numVideo]).Moniker;

  if (numAudio>=0) and(numAudio<AudioDevicesList.Count) then
    nmAudio:= TCapDeviceInfo(AudioDevicesList[numAudio]).Moniker;

  ChooseDevices(nmVideo, nmAudio, ForceRebuild);
  nmVideo:= nil; nmAudio:= nil;
end;

//-----------------------------------------------------------------

procedure TVideoCapture.ChooseCompressors(nmVComp, nmAComp: IMoniker; ForceRebuild: boolean);
begin
  if (ForceRebuild) or((mVComp<>nmVComp) or (mAComp<>nmAComp)) then  begin
    mVComp:= nil; mAComp:= nil;
    mVComp:= nmVComp; mAComp:= nmAComp;
    TearDownGraph;
    BuildGraph;
    if FWantPreview then StartPreview;
    if Assigned(FOnChangeCompressor) then FOnChangeCompressor(Self);
  end;
end;

procedure TVideoCapture.ChooseCompressors(const numVComp, numAComp: integer; ForceRebuild: boolean);
var
  nmVComp, nmAComp: IMoniker;
begin
  if (numVComp>=0) and(numVComp<VideoCompressorsList.Count) then
    nmVComp:= TCapDeviceInfo(VideoCompressorsList[numVComp]).Moniker;

  if (numAComp>=0) and(numAComp<AudioCompressorsList.Count) then
    nmAComp:= TCapDeviceInfo(AudioCompressorsList[numAComp]).Moniker;

  ChooseCompressors(nmVComp, nmAComp, ForceRebuild);
  nmVComp:= nil; nmAComp:= nil;
end;

procedure TVideoCapture.ChooseCompressors(const szVComp, szAComp: string; ForceRebuild: boolean);
var
  nmVComp, nmAComp: IMoniker;
  i: integer;
begin

  for i:= 0 to VideoCompressorsList.Count-1 do
    with VideoCompressorsList[i] as TCapDeviceInfo do
      if DeviceNameInList = szVComp then begin
        nmVComp:= Moniker;
        break;
      end;

  for i:= 0 to AudioCompressorsList.Count-1 do
    with AudioCompressorsList[i] as TCapDeviceInfo do
      if DeviceNameInList = szAComp then begin
        nmAComp:= Moniker;
        break;
      end;


  ChooseCompressors(nmVComp, nmAComp, ForceRebuild);
  nmVComp:= nil; nmAComp:= nil;
end;

//-----------------------------------------------------------------

procedure TVideoCapture.CleanUp;
begin
  FreeCapFilters;

  VideoWindow:= nil;
  MediaEvent:= nil;
  DroppedFrames:= nil;

  Render:= nil;
  Sink:= nil;
  ConfigAviMux:= nil;

  fCapAudioIsRelevant:= false;
  fCapAudio:= false;
  fCCAvail:= false;
  fCapCC:= false;
  fGraphBuilt:= false;
  fPreviewFaked:= false;
  fCapturing:= false;
  fPreviewing:= false;
  FMasterStream:= -1;
end;

procedure TVideoCapture.FreeCapFilters;
begin
  Graph:= nil;
  Builder:= nil;
  VCap:= nil;
  ACap:= nil;
  VComp:= nil;
  AComp:= nil;
  AStreamConf:= nil;
  VStreamConf:= nil;
  VideoCompression:= nil;
  CaptureDialogs:= nil;
  Grabber:= nil;
  FDialogs:= [];
end;

function TVideoCapture.MakeBuilder: boolean;
begin
  Result:= (Builder <> nil) or
           (CoCreateInstance( CLSID_CaptureGraphBuilder2, nil, CLSCTX_INPROC,
                              IID_ICaptureGraphBuilder2, Builder) = NOERROR);
end;

function TVideoCapture.MakeGraph: boolean;
begin
  Result:= (Graph <> nil) or
           (CoCreateInstance( CLSID_FilterGraph, nil, CLSCTX_INPROC,
			      IID_IGraphBuilder, Graph) = NOERROR);
end;

function TVideoCapture.InitCapFilters: boolean;
var
  PropBag: IPropertyBag;
  hr: HRESULT;
  varOle: OleVariant;
  pmt: PAM_MEDIA_TYPE;
  Pin: IPin;
  pins: IEnumPins;
  n: ULONG;
  pinInfo: TPIN_INFO;
  Found: boolean;
  Ks: IKsPropertySet;
  guid: TGUID;
  dw: DWORD;
begin
  hr:= 0;
  Result:= MakeBuilder;
  if not Result then begin
    ErrMsg(rsCantMakeGraphBuilder);
    Exit;
  end;

  try
    VCap:= nil;
    if mVideo <> nil then begin
      hr:= mVideo.BindToStorage(nil, nil, IID_IPropertyBag, PropBag);
      if Succeeded(hr) then begin
        if PropBag.Read('FriendlyName', varOle, nil)=NOERROR then FVCapFriendlyName:= varOle;
        PropBag:= nil;
      end;
      hr:= mVideo.BindToObject(nil, nil, IID_IBaseFilter, VCap);
    end;
    if VCap = nil then
      ErrMsgException(rsCantCreateVCaptureFilter, hr);

    // make a filtergraph, give it to the graph builder and put the video
    // capture filter in the graph
    if not MakeGraph then ErrMsgException(rsCantMakeGraph);

    if Builder.SetFiltergraph(Graph) <> NOERROR then
      ErrMsgException(rsCantSetFilterGraph);

    if Graph.AddFilter(VCap, nil) <> NOERROR then
      ErrMsgException(rsCantAddVFilterToGraph, hr);

    // Calling FindInterface below will result in building the upstream
    // section of the capture graph (any WDM TVTuners or Crossbars we might need).
    // we use this interface to get the name of the driver
    // Don't worry if it doesn't work:  This interface may not be available
    // until the pin is connected, or it may not be available at all.
    // (eg: interface may not be available for some DV capture)
    hr:= Builder.FindInterface( @PIN_CATEGORY_CAPTURE, @MEDIATYPE_Interleaved,
                                VCap, IID_IAMVideoCompression, VideoCompression);
    if hr <> S_OK then
      Builder.FindInterface( @PIN_CATEGORY_CAPTURE, @MEDIATYPE_Video,
                             VCap, IID_IAMVideoCompression, VideoCompression);

    // !!! What if this interface isn't supported?
    // we use this interface to set the frame rate and get the capture size
    hr:= Builder.FindInterface( @PIN_CATEGORY_CAPTURE, @MEDIATYPE_Interleaved,
                                VCap, IID_IAMStreamConfig, VStreamConf);
    if hr <> NOERROR then begin
      hr:= Builder.FindInterface( @PIN_CATEGORY_CAPTURE, @MEDIATYPE_Video,
                                  VCap, IID_IAMStreamConfig, VStreamConf);
      if hr <> NOERROR then // this means we can't set frame rate (non-DV only)
        ErrMsg(rsCantFindVStreamConfig, hr);
    end;

    fCapAudioIsRelevant:= true;
    fCapAudio:= FWantAudio;

    if (VStreamConf <> nil)  then begin
      // default capture format
      if (VStreamConf.GetFormat(pmt)=S_OK) then
        try
          // DV capture does not use a VIDEOINFOHEADER
          if CheckGUID(pmt^.formattype, FORMAT_VideoInfo) then begin
            // resize our window to the default capture size
            FCaptureWidth:= ABS(PVIDEOINFOHEADER(pmt^.pbFormat)^.bmiHeader.biWidth);
            FCaptureHeight:= ABS(PVIDEOINFOHEADER(pmt^.pbFormat)^.bmiHeader.biHeight);
            ResizeWindow;
          end;
          if not CheckGUID(pmt^.majortype, MEDIATYPE_Video) then begin
            // This capture filter captures something other that pure video.
            // Maybe it's DV or something?  Anyway, chances are we shouldn't
            // allow capturing audio separately, since our video capture
            // filter may have audio combined in it already!
            fCapAudioIsRelevant:= false;
            fCapAudio:= false;
          end;
        finally
          DeleteMediaType(pmt);
        end;
    end;
    // we use this interface to bring up the 3 dialogs
    // NOTE:  Only the VfW capture filter supports this.  This app only brings
    // up dialogs for legacy VfW capture drivers, since only those have dialogs
    Builder.FindInterface( @PIN_CATEGORY_CAPTURE, @MEDIATYPE_Video,
                           VCap, IID_IAMVfwCaptureDialogs, CaptureDialogs);
    Found:= false;
    Pin:= nil;
    if Succeeded(VCap.EnumPins(pins)) then begin
      while (not Found) and (pins.Next(1, pin, @n) = S_OK) do begin
        if S_OK = pin.QueryPinInfo(pinInfo) then begin
          if pinInfo.dir = PINDIR_INPUT then begin
            // is this pin an ANALOGVIDEOIN input pin?
            if pin.QueryInterface(IID_IKsPropertySet, Ks) = S_OK then begin
              Found:= (Ks.Get(AMPROPSETID_Pin, 0, nil, 0, @guid, sizeof(guid), dw) = S_OK) and
                       CheckGuid(guid, PIN_CATEGORY_ANALOGVIDEOIN);
              Ks:= nil;
            end;
          end;
          pinInfo.pFilter:= nil;
        end;
        pin:= nil;
      end;
      pins:= nil;
    end;

    // there's no point making an audio capture filter
    if (FWantAudio) and (fCapAudioIsRelevant) then begin
      // create the audio capture filter, even if we are not capturing audio right
      // now, so we have all the filters around all the time.
      //
      // We want an audio capture filter and some interfaces
      if (mAudio = nil) then begin
        // there are no audio capture devices. We'll only allow video capture
        fCapAudio:= false;
      end
      else begin
        ACap:= nil;
        hr:= mAudio.BindToStorage(nil, nil, IID_IPropertyBag, PropBag);
        if Succeeded(hr) then begin
          if PropBag.Read('FriendlyName', varOle, nil)=NOERROR then FACapFriendlyName:= varOle;
          PropBag:= nil;
        end;

        mAudio.BindToObject(nil, nil, IID_IBaseFilter, ACap);
        if ACap = nil then begin
          // there are no audio capture devices. We'll only allow video capture
          fCapAudio:= false;
          ErrMsg(rsCantMakeACapFilter);
        end
        else begin
          // put the audio capture filter in the graph
          // We'll need this in the graph to get audio property pages
          if Graph.AddFilter(ACap, nil) <> NOERROR then
            ErrMsgException(rsCantAddAFilterToGraph, hr);

          // Calling FindInterface below will result in building the upstream
          // section of the capture graph (any WDM TVAudio's or Crossbars we might need).

          // !!! What if this interface isn't supported?
          // we use this interface to set the captured wave format
          hr:= Builder.FindInterface( @PIN_CATEGORY_CAPTURE, @MEDIATYPE_Audio,
                                      ACap, IID_IAMStreamConfig, AStreamConf );
          if hr <> NOERROR then ErrMsg(rsCantFindAStreamConfig);
        end;
      end;
    end;

    // Can this filter do closed captioning?
    hr:= Builder.FindPin( VCap, PINDIR_OUTPUT, PIN_CATEGORY_VBI, TGUID(nil^),
                          false, 0, Pin);
    if hr <> S_OK then
      hr:= Builder.FindPin( VCap, PINDIR_OUTPUT, PIN_CATEGORY_CC, TGUID(nil^),
                            false, 0, Pin);
    fCCAvail:= (hr = S_OK); // can't capture it, then
    if fCapCC then Pin:= nil;

    GetAvailableDialogs;
    Result:= true;
  except
    FreeCapFilters;
    Result:= false;
  end;
end;

function TVideoCapture.Init: boolean;
begin
  // Create the filter graph and create the capture graph builder.
  Result:= MakeGraph and MakeBuilder;
  if not Result then Exit;

  Builder.SetFiltergraph(Graph);
  BuildCaptureDevicesList;
  Result:= (VideoDevicesList.Count>0) or (AudioDevicesList.Count>0);
end;

procedure TVideoCapture.ResizeWindow;
begin
  if Assigned(VideoWindow) then
    VideoWindow.SetWindowPosition(0, 0, ClientWidth, ClientHeight);
end;

procedure TVideoCapture.SetSize(var msg: TMessage);
begin
  inherited;
  ResizeWindow;
end;

// graph event occured
// get events
procedure TVideoCapture.GraphEvent(var msg: TMessage);
var
  Event, l1, l2: integer;
  wasCapturing, wasPreviewing: boolean;
begin
  wasCapturing:= Capturing;
  wasPreviewing:= Previewing;
  if (MediaEvent <> nil) then begin
    while MediaEvent.GetEvent(Event, l1, l2, 0) = S_OK do
      try
        case Event of
          EC_ERRORABORT: StopCapture;
          EC_DEVICE_LOST: begin
                            StopCapture;
                            StopPreview;
                          end;
        end;
      finally
        MediaEvent.FreeEventParams(Event, l1, l2);
      end;
    // we have stopped capture now need to restore preview
    if ((Capturing<>wasCapturing) or (wasPreviewing<>Previewing)) and (FWantPreview) then begin
      BuildGraph;
      StartPreview;
    end;
  end;
end;

function TVideoCapture.RenderPreviewPin: boolean;
var
  hr: HRESULT;
  DVDec: IBaseFilter;
  DVDecProp: IIPDVDec;
begin
  Result:= true;
  fPreviewFaked:= false;
  // Render the preview pin - even if there is not preview pin, the capture
  // graph builder will use a smart tee filter and provide a preview.
  // !!! what about latency/buffer issues?
  // NOTE that we try to render the interleaved pin before the video pin, because
  // if BOTH exist, it's a DV filter and the only way to get the audio is to use
  // the interleaved pin.  Using the Video pin on a DV filter is only useful if
  // you don't want the audio.
  hr:= Builder.RenderStream(@PIN_CATEGORY_PREVIEW, @MEDIATYPE_Interleaved, VCap, nil, nil);
  if hr = VFW_S_NOPREVIEWPIN then begin
    // preview was faked up for us using the (only) capture pin
    fPreviewFaked:= true;
  end
  else if hr <> S_OK then begin
    // maybe it's DV?
    hr:= Builder.RenderStream(@PIN_CATEGORY_PREVIEW, @MEDIATYPE_Video, VCap, nil, nil);
    if hr = VFW_S_NOPREVIEWPIN then begin
      // preview was faked up for us using the (only) capture pin
      fPreviewFaked:= true;
    end
    else
      Result:= hr = S_OK;
  end;
  if Result then begin
    if (FDVSize<>dvsDontWorry) and
       (Graph.FindFilterByName('DV Video Decoder', DVDec) =S_OK) and
       (DVDec.QueryInterface(IIPDVDec, DVDecProp)=S_OK) then
    begin
      DVDecProp.put_IPDisplay(DVSizes[FDVSize]);
    end;
  end;
end;

function FindPin(pins: IEnumPins; PinDir: TPin_Direction): IPin;
var
  pin: IPin;
  pininfo: TPin_Direction;
  hr: HRESULT;
  u: ULONG;
begin
  Result:= nil;
  if pins=nil then Exit;
  pins.Reset;
  repeat
    hr:= pins.Next(1, pin, @u);
    if (hr = S_OK) then begin
      hr:= pin.QueryDirection(pininfo);
      if (hr = S_OK) and (pininfo = PinDir) then begin
        Result:= pin;
        break;
      end;
    end;
  until hr<>NOERROR;
end;

const
  PixelFormatGUIDs: array [TPixelFormat] of PGUID = (
          @MEDIASUBTYPE_RGB24,
          @MEDIASUBTYPE_RGB1,
          @MEDIASUBTYPE_RGB4,
          @MEDIASUBTYPE_RGB8,
          @MEDIASUBTYPE_RGB555,
          @MEDIASUBTYPE_RGB565,
          @MEDIASUBTYPE_RGB24,
          @MEDIASUBTYPE_RGB32,
          @MEDIASUBTYPE_RGB24
  );

// we have built preview graph
// now we have to find video renderer to
// paste grabber before renderer's position and
// then reconnect video renderer again by rendering grabber's output
function TVideoCapture.AdjustGrabberPosition: boolean;
var
  FGrabber, FRenderer: IBaseFilter;
  GrabIn, GrabOut, RenderIn, DecoderOut: IPin;
  pins: IEnumPins;
  mt: TAM_Media_Type;
  hr: HRESULT;
begin
  // find video renderer to put grabber just before it
  // I dont know how to find it another way except by Name(((((((
  Result:= FWantBitmaps and (Graph.FindFilterByName('Video Renderer', FRenderer) =S_OK);
  if Result then begin
    Result:= (CoCreateInstance(CLSID_SampleGrabber, nil, CLSCTX_INPROC, IID_IBaseFilter, FGrabber) = NOERROR) and
             (FGrabber.QueryInterface(IID_ISampleGrabber, Grabber) = S_OK) and
             (Graph.AddFilter(FGrabber, 'Grabber') = S_OK) and
             (FGrabber.EnumPins(pins)=S_OK);
    if (Result) then begin
      GrabIn:= FindPin(pins, PINDIR_INPUT);
      GrabOut:= FindPin(pins, PINDIR_OUTPUT);
      pins:= nil;
      if (FRenderer.EnumPins(pins)=S_OK) then begin
        RenderIn:= FindPin(pins, PINDIR_INPUT);
        RenderIn.ConnectedTo(DecoderOut);
        pins:= nil;
        // just disconnect the render not remove it
        hr:= Graph.Disconnect(RenderIn);
        Result:= (hr = S_OK) or (hr = S_FALSE);
        hr:= Graph.Disconnect(DecoderOut);
        Result:= Result and ((hr = S_OK) or (hr = S_FALSE));
        if Result then begin
          ZeroMemory(@mt, sizeof(mt));
          mt.formattype:= FORMAT_VideoInfo;
          mt.majortype:= MEDIATYPE_Video;
          mt.subtype:= PixelFormatGUIDs[FPixelFormat]^;
          Grabber.SetMediaType(mt);
          Result:= (DecoderOut<>nil) and (GrabIn<>nil) and
                   (Graph.Connect(DecoderOut, GrabIn)=S_OK);

          if not Result then begin
            // we cant connect grabber - probably wrong mediatype?
            // TO DO: try to request DecoderOut available mediatypes and
            //        set one of them
          end;
          if Result then begin
            Grabber.SetBufferSamples(false);
            Grabber.SetOneShot(false);
            TSampleGrabberCB(GrabberCB).Owner:= Handle;
            Grabber.SetCallback(TSampleGrabberCB(GrabberCB), 1);
            Result:= Graph.Connect(GrabOut, RenderIn) = S_OK ;
          end
          else begin
            // we cant put or connect grabber - need to restore preview
            Grabber:= nil;
            Result:= Graph.Connect(DecoderOut, RenderIn) = S_OK;
            ErrMsg(rsCantInstallSampleGrabber);
          end;
        end;
      end;
    end;
  end;
end;

function TVideoCapture.FindVideoWindow: boolean;
begin
  // Get the preview window to be a child of our app's window
  // This will find the IVideoWindow interface on the renderer.  It is
  // important to ask the filtergraph for this interface... do NOT use
  // ICaptureGraphBuilder2::FindInterface, because the filtergraph needs to
  // know we own the window so it can give us display changed messages, etc.
  Result:= (Graph.QueryInterface(IID_IVideoWindow, VideoWindow)=S_OK);
  if Result then begin
    VideoWindow.put_Owner(Handle);    // We own the window now
    VideoWindow.put_WindowStyle(WS_CHILD or WS_CLIPSIBLINGS);    // you are now a child
    // give the preview window all our space
    ResizeWindow;
    VideoWindow.put_Visible(true);
  end;
end;

// Tear down everything downstream of a given filter
procedure TVideoCapture.NukeDownstream(pf: IBaseFilter);
var
  pP, pTo: IPin;
  u: ULONG;
  pins: IEnumPins;
  pininfo: TPIN_INFO;
  hr: HRESULT;
begin
  pins:= nil;
  hr:= pf.EnumPins(pins);
  pins.Reset;
  while hr = NOERROR do begin
    hr:= pins.Next(1, pP, @u);
    if (hr = S_OK) and (pP <> nil) then begin
      pP.ConnectedTo(pTo);
      if pTo <> nil then begin
        hr:= pTo.QueryPinInfo(pininfo);
        if hr = NOERROR then begin
  	  if pininfo.dir = PINDIR_INPUT then begin
	    NukeDownstream(pininfo.pFilter);
	    Graph.Disconnect(pTo);
	    Graph.Disconnect(pP);
            Graph.RemoveFilter(pininfo.pFilter);
	  end;
          pininfo.pFilter:= nil;
        end;
        pTo:= nil;
      end;
      pP:= nil;
    end;
  end;
  pins:= nil;
end;

// Tear down everything downstream of the capture filters, so we can build
// a different capture graph.  Notice that we never destroy the capture filters
// and WDM filters upstream of them, because then all the capture settings
// we've set would be lost.
procedure TVideoCapture.TearDownGraph;
begin
  if FCapturing then StopCapture;
  if FPreviewing then StopPreview;
  Sink:= nil;
  ConfigAviMux:= nil;
  Render:= nil;
  if VideoWindow <> nil then begin
    // stop drawing in our window, or we may get wierd repaint effects
    VideoWindow.put_Visible(false);
    VideoWindow.put_Owner(0);
  end;
  VideoWindow:= nil;
  MediaEvent:= nil;
  DroppedFrames:= nil;

  // free compressors cause we have to insert during graph building
  VComp:= nil;
  AComp:= nil;

  // destroy the graph downstream of our capture filters
  if VCap <> nil then NukeDownstream(VCap);
  if ACap <> nil then NukeDownstream(ACap);

  fGraphBuilt:= false;
  fPreviewFaked:= false;
  if (FWantCapture) then DeleteFile(TempCaptureFileName);
end;

// build the preview graph!
// !!! PLEASE NOTE !!!  Some new WDM devices have totally separate capture
// and preview settings.  An application that wishes to preview and then
// capture may have to set the preview pin format using IAMStreamConfig on the
// preview pin, and then again on the capture pin to capture with that format.
// In this sample app, there is a separate page to set the settings on the
// capture pin and one for the preview pin.  To avoid the user
// having to enter the same settings in 2 dialog boxes, an app can have its own
// UI for choosing a format (the possible formats can be enumerated using
// IAMStreamConfig) and then the app can programmatically call IAMStreamConfig
// to set the format on both pins.

// build the capture graph!
function TVideoCapture.BuildGraph: boolean;
var
   hr: HRESULT;
   pmt: PAM_MEDIA_TYPE;
begin
  // we have one already
  Result:= fGraphBuilt;
  if Result then Exit;

  // We don't have the necessary capture filters
  if (VCap = nil) then Exit;
  if (ACap = nil) and (FWantAudio) then Exit;

  if (FWantCapture) then begin
    // no capture file name yet... we need one first
    if UseTempFile then
      FCaptureFile:= TempCaptureFileName
    else
      FCaptureFile:= CaptureFileName;

    if (FCaptureFile = '') then begin
      ErrMsg(rsEmptyFileName);
      Result:= false;
      Exit;
    end;

    DeleteFile(FCaptureFile);

    if FDoPreallocFile and not AllocCaptureFile(PreallocFileSize) then begin
      ErrMsg(rsFailedToAllocFileSize);
      Result:= false;
      Exit;
    end;
  end;

  try
// We need a rendering section that will write the capture file out in AVI file format
    if (FWantCapture) then  begin
      hr:= Builder.SetOutputFileName(MEDIASUBTYPE_Avi, PWCHAR(FCaptureFile), Render, Sink);
      if (hr <> NOERROR) then ErrMsgException(rsCantSetCaptureFile);

  // Now tell the AVIMUX to write out AVI files that old apps can read properly.
  // If we don't, most apps won't be able to tell where the keyframes are, slowing down editing considerably
  // Doing this will cause one seek (over the area the index will go) when
  // you capture past 1 Gig, but that's no big deal.
  // NOTE: This is on by default, so it's not necessary to turn it on
      hr:= Render.QueryInterface(IID_IConfigAviMux, ConfigAviMux);
      if (hr = NOERROR) and (ConfigAviMux<>nil) then begin
        ConfigAviMux.SetOutputCompatibilityIndex(true);
        if (fCapAudio) then // Also, set the proper MASTER STREAM
          MasterStream:= FMasterStream;
      end;

      // adding video compression filter
      if mVComp <> nil then begin
         VComp:= nil;
         hr:= mVComp.BindToObject(nil, nil, IID_IBaseFilter, VComp);
         if (hr=S_OK) and Failed(Graph.AddFilter(VComp, VCompFilterName)) then
           ErrMsg(rsCantAddVCompFilter);
      end;
// Render the video capture and preview pins - even if the capture filter only
// has a capture pin (and no preview pin) this should work... because the
// capture graph builder will use a smart tee filter to provide both capture
// and preview.  We don't have to worry.  It will just work.

// NOTE that we try to render the interleaved pin before the video pin, because
// if BOTH exist, it's a DV filter and the only way to get the audio is to use
// the interleaved pin.  Using the Video pin on a DV filter is only useful if
// you don't want the audio.
      hr:= Builder.RenderStream( @PIN_CATEGORY_CAPTURE, @MEDIATYPE_Interleaved,
                                 VCap, VComp, Render);
      if (hr <> NOERROR) then begin
          hr:= Builder.RenderStream( @PIN_CATEGORY_CAPTURE, @MEDIATYPE_Video,
                                     VCap, VComp, Render);
          if (hr <> NOERROR) then ErrMsgException(rsCantRenderVCaptureStream);
      end;
    end;

    // Rendering preview pin
    if (FWantPreview) then begin
       if RenderPreviewPin then
          AdjustGrabberPosition()
       else
          ErrMsgException(rsCantRenderPreviewStream);
    end;

    if (fCapAudio) then begin // Render the audio capture pin?
      // adding audio compression filter
      if (FWantAudio) and (mAComp <> nil) then begin
         AComp:= nil;
         hr:= mAComp.BindToObject(nil, nil, IID_IBaseFilter, AComp);
         if (hr=S_OK) and Failed(Graph.AddFilter(AComp, ACompFilterName)) then
           ErrMsg(rsCantAddACompFilter);
      end;

      hr:= Builder.RenderStream( @PIN_CATEGORY_CAPTURE, @MEDIATYPE_Audio,
                                 ACap, AComp, Render);
      if (hr <> NOERROR) then
        ErrMsgException(rsCantRenderACaptureStream);
    end;

    // Render the closed captioning pin? It could be a CC or a VBI category pin,
    // depending on the capture driver
    if (fCapCC) then begin
      hr:= Builder.RenderStream(@PIN_CATEGORY_CC, nil, VCap, nil, Render);
      if (hr <> NOERROR) then begin
        hr:= Builder.RenderStream(@PIN_CATEGORY_VBI, nil, VCap, nil, Render);
        if (hr <> NOERROR) then ErrMsg(rsCantRenderCC);
      end;
      // To preview and capture VBI at the same time, we can call this twice
      if (fWantPreview) then
        Builder.RenderStream(@PIN_CATEGORY_VBI, nil, VCap, nil, nil);
    end;

    // NOTE: We do this even if we didn't ask for a preview, because rendering
    // the capture pin may have rendered the preview pin too (WDM overlay
    // devices) because they must have a preview going.  So we better always
    // put the preview window in our app, or we may get a top level window
    // appearing out of nowhere!
    if not FindVideoWindow and (fWantPreview) then ErrMsg(rsThisGraphCantPreview);

    // now tell it what frame rate to capture at.  Just find the format it
    // is capturing with, and leave everything alone but change the frame rate
    if fUseFrameRate then
      hr:= E_FAIL
    else
      hr:= NOERROR;

    if (VStreamConf<>nil) and fUseFrameRate then begin
	// DV capture does not use a VIDEOINFOHEADER
        if VStreamConf.GetFormat(pmt) = NOERROR then begin
          try
            if CheckGuid(pmt^.formattype, FORMAT_VideoInfo) then begin
              PVIDEOINFOHEADER(pmt^.pbFormat)^.AvgTimePerFrame:= round(10000000 / FrameRate);
              hr:= VStreamConf.SetFormat(pmt^);
              if hr <> NOERROR then ErrMsg(rsCantSetPreviewFrameRate, hr);
            end;
          finally
            DeleteMediaType(pmt);
          end;
        end;
    end;
    if (hr <> NOERROR) then ErrMsg(rsCantSetCaptureFrameRate);

    // now ask the filtergraph to tell us when something is completed or aborted
    // (EC_COMPLETE, EC_USERABORT, EC_ERRORABORT).  This is how we will find out
    // if the disk gets full while capturing
    if (Graph.QueryInterface(IID_IMediaEventEx, MediaEvent) = NOERROR) then
      MediaEvent.SetNotifyWindow(Handle, WM_FGNOTIFY, 0);

// All done.
    fGraphBuilt:= true;
    Result:= true;
  except
    TearDownGraph;
    Result:= false;
  end;

end;

//--------------------------------------------------------------------
// working with avaiable modes of current capture filters
function TVideoCapture.GetACapModeCount: integer;
var
  capCount, capSize: integer;
begin
  Result:= 0;
  if (AStreamConf<>nil) and (AStreamConf.GetNumberOfCapabilities(capCount, capSize)=S_OK) then begin
    Result:= capCount;
  end;
end;

function TVideoCapture.GetACapMode: TACapMode;
var
  pmt: PAM_MEDIA_TYPE;
begin
  ZeroMemory(@Result, SizeOf(Result));
  if (AStreamConf<>nil) and (AStreamConf.GetFormat(pmt)=S_OK) then
    try
      if CheckGUID(pmt^.formattype, FORMAT_WaveFormatEx) then begin
        Result.MediaType:= pmt^.majortype;
        Result.MediaSubType:= pmt^.subtype;
        Result.SampleRate:= PWAVEFORMATEX(pmt^.pbFormat)^.nSamplesPerSec;
        Result.SampleSize:= PWAVEFORMATEX(pmt^.pbFormat)^.wBitsPerSample;
        Result.Channels:= PWAVEFORMATEX(pmt^.pbFormat)^.nChannels;
      end;
    finally
      DeleteMediaType(pmt);
    end;
end;

function TVideoCapture.GetACapModes(i: integer): TACapMode;
var
  pmt: PAM_MEDIA_TYPE;
  aConfig: TAudio_Stream_Config_Caps;
begin
  ZeroMemory(@Result, SizeOf(Result));
  if (i>=0) and (i<ACapModeCount) and (AStreamConf.GetStreamCaps(i, pmt, aConfig)=S_OK) then
    try
      if CheckGUID(pmt^.formattype, FORMAT_WaveFormatEx) then begin
        Result.MediaType:= pmt^.majortype;
        Result.MediaSubType:= pmt^.subtype;
        Result.SampleRate:= PWAVEFORMATEX(pmt^.pbFormat)^.nSamplesPerSec;
        Result.SampleSize:= PWAVEFORMATEX(pmt^.pbFormat)^.wBitsPerSample;
        Result.Channels:= PWAVEFORMATEX(pmt^.pbFormat)^.nChannels;
      end;
    finally
      DeleteMediaType(pmt);
    end;
end;

function TVideoCapture.SetACapMode(i: integer): boolean;
var
  pmt: PAM_MEDIA_TYPE;
  aConfig: TAudio_Stream_Config_Caps;
begin
  Result:= false;
  if (i>=0) and (i<ACapModeCount) and (AStreamConf.GetStreamCaps(i, pmt, aConfig)=S_OK) then
    try
      // we need to stop streaming first and then recconect pin?
      TearDownGraph();
      Result:= AStreamConf.SetFormat(pmt^)=S_OK;
      if FWantPreview then begin
        BuildGraph;
        StartPreview();
      end;
    finally
      DeleteMediaType(pmt);
    end;
  if Result and Assigned(FOnChangeACapMode) then FOnChangeACapMode(Self);
end;

function TVideoCapture.SetACapMode(const mode: TACapMode): boolean;
var
  i, capCount: integer;
begin
  Result:= IsEqualModes(mode, ACapMode);
  if not Result then begin
    capCount:= ACapModeCount;
    for i:= 0 to capCount-1 do
      if IsEqualModes(mode, ACapModes[i]) then begin
        Result:= SetACapMode(i);
        break;
      end;
  end;
end;

function TVideoCapture.GetVCapModeCount: integer;
var
  capCount, capSize: integer;
begin
  Result:= 0;
  if (VStreamConf<>nil) and (VStreamConf.GetNumberOfCapabilities(capCount, capSize)=S_OK) then begin
    Result:= capCount;
  end;
end;

function TVideoCapture.GetVCapMode: TVCapMode;
var
  pmt: PAM_MEDIA_TYPE;
begin
  ZeroMemory(@Result, SizeOf(Result));
  if (VStreamConf<>nil) and (VStreamConf.GetFormat(pmt)=S_OK) then
    try
      if CheckGUID(pmt^.formattype, FORMAT_VideoInfo) then begin
        Result.MediaType:= pmt^.majortype;
        Result.MediaSubType:= pmt^.subtype;
        Result.Width:= ABS(PVIDEOINFOHEADER(pmt^.pbFormat)^.bmiHeader.biWidth);
        Result.Height:= ABS(PVIDEOINFOHEADER(pmt^.pbFormat)^.bmiHeader.biHeight);
        Result.BitCount:= PVIDEOINFOHEADER(pmt^.pbFormat)^.bmiHeader.biBitCount;
      end;
    finally
      DeleteMediaType(pmt);
    end;
end;

function TVideoCapture.GetVCapModes(i: integer): TVCapMode;
var
  pmt: PAM_MEDIA_TYPE;
  vConfig: TVideo_Stream_Config_Caps;
begin
  ZeroMemory(@Result, SizeOf(Result));
  if (i>=0) and (i<VCapModeCount) and (VStreamConf.GetStreamCaps(i, pmt, vConfig)=S_OK) then
    try
      if CheckGUID(pmt^.formattype, FORMAT_VideoInfo) then begin
        Result.MediaType:= pmt^.majortype;
        Result.MediaSubType:= pmt^.subtype;
        Result.Width:= ABS(PVIDEOINFOHEADER(pmt^.pbFormat)^.bmiHeader.biWidth);
        Result.Height:= ABS(PVIDEOINFOHEADER(pmt^.pbFormat)^.bmiHeader.biHeight);
        Result.BitCount:= PVIDEOINFOHEADER(pmt^.pbFormat)^.bmiHeader.biBitCount;
        if (vConfig.MaxFrameInterval>0) then
          Result.MinFrameRate:= 10000000 / vConfig.MaxFrameInterval; //??
        if (vConfig.MinFrameInterval>0) then
          Result.MaxFrameRate:= 10000000 / vConfig.MinFrameInterval; //??
        if (PVIDEOINFOHEADER(pmt^.pbFormat)^.AvgTimePerFrame>0) then
          Result.FrameRate:= 10000000 / PVIDEOINFOHEADER(pmt^.pbFormat)^.AvgTimePerFrame; //??
      end;
    finally
      DeleteMediaType(pmt);
    end;
end;

function TVideoCapture.SetVCapMode(i: integer): boolean;
var
  pmt: PAM_MEDIA_TYPE;
  vConfig: TVideo_Stream_Config_Caps;
begin
  Result:= false;
  if FCapturing then Exit;  // dont let change format during capture
  if (i>=0) and (i<VCapModeCount) and (VStreamConf.GetStreamCaps(i, pmt, vConfig)=S_OK) then
    try
      // we need to stop streaming first and then recconect pin?
      TearDownGraph();
      Result:= VStreamConf.SetFormat(pmt^)=S_OK;
      if FWantPreview then begin
        BuildGraph;
        StartPreview();
      end;
    finally
      DeleteMediaType(pmt);
    end;
  if Result and Assigned(FOnChangeVCapMode) then FOnChangeVCapMode(Self);
end;

function TVideoCapture.SetVCapMode(const mode: TVCapMode): boolean;
var
  i, capCount: integer;
begin
  Result:= IsEqualModes(mode, VCapMode);
  if not Result then begin
    capCount:= VCapModeCount;
    for i:= 0 to capCount-1 do
      if IsEqualModes(mode, VCapModes[i]) then begin
        Result:= SetVCapMode(i);
        break;
      end;
  end;
end;

//--------------------------------------------------------------------

procedure TVideoCapture.SetMasterStream(const Value: integer);
begin
  if (ConfigAviMux<>nil) and (ConfigAviMux.SetMasterStream(Value) = NOERROR) then
    FMasterStream:= Value
  else
    ErrMsg(rsSetMasterStreamFailed);
end;

// capture single frame
procedure TVideoCapture.BitmapGrabbed(var msg: TMessage);
var
  MediaType: TAM_Media_Type;
  bih: TBitmapInfoHeader;
  BitmapInfo: TBITMAPINFO;
  w, h: integer;
  usage: cardinal;
  CapturedImage: TBitmap;
begin
  if (TSampleGrabberCB(GrabberCB).BufferLength>0) and
     (Grabber.GetConnectedMediaType(MediaType) = S_OK) then
  begin
    try
      // Get a pointer to the video header.
      if CheckGUID(MediaType.FormatType, FORMAT_VideoInfo) then
        bih:= PVideoInfoHeader(mediatype.pbFormat)^.bmiHeader
      else if CheckGUID(MediaType.FormatType, FORMAT_VideoInfo2) then
        bih:= PVideoInfoHeader2(mediatype.pbFormat)^.bmiHeader
      else
        Abort;
      // The video header contains the bitmap information.
      // Copy it into a BITMAPINFO structure.
      ZeroMemory(@BitmapInfo, sizeof(BitmapInfo));
      Move(bih, BitmapInfo.bmiHeader, sizeof(bih));
    finally
      FreeMediaType(MediaType);
    end;
    // Copy the image into the buffer.
    w:= BitmapInfo.bmiHeader.biWidth;
    h:= BitmapInfo.bmiHeader.biHeight;
    CapturedImage:= TBitmap.Create;
    try
      case BitmapInfo.bmiHeader.biBitCount of
         1: CapturedImage.PixelFormat:= pf1bit;
         4: CapturedImage.PixelFormat:= pf4bit;
         8: CapturedImage.PixelFormat:= pf8bit;
        15: CapturedImage.PixelFormat:= pf15bit;
        16: CapturedImage.PixelFormat:= pf16bit;
        24: CapturedImage.PixelFormat:= pf24bit;
        32: CapturedImage.PixelFormat:= pf32bit;
      else
        // strange color depth - let bitmap be unchanged
      end;
      CapturedImage.Width:= w;  // enlarge image to ensure the whole DIB gets copied
      CapturedImage.Height:= h;

      if (BitmapInfo.bmiHeader.biClrUsed=0) then
        usage:= DIB_RGB_COLORS  // BITMAP contains real color values
      else
        usage:= DIB_PAL_COLORS; // BITMAP uses pallete

      if (StretchDIBits( CapturedImage.Canvas.Handle, // handle of device context
                         0, 0, w, h,                  // destination rectangle
                         0, 0, w, h,                  // source rectangle
                         TSampleGrabberCB(GrabberCB).Buffer, // address of bitmap bits
                         BitmapInfo,	             // address of bitmap data
                         usage, SRCCOPY ) <> 0 )
      then begin
        if Assigned(FOnBitmapGrabbed) then FOnBitmapGrabbed(CapturedImage);
      end;
    finally
      CapturedImage.Free;
    end;
    TSampleGrabberCB(GrabberCB).ClearBuffer;
  end;
end;

function TVideoCapture.CaptureFrame: boolean;
begin
  Result:= (Grabber<>nil) and FWantBitmaps and (Previewing or Capturing);
  TSampleGrabberCB(GrabberCB).GrabbingEnabled:= Result;
end;

function TVideoCapture.StartPreview: boolean;
var
  MC: IMediaControl;
  hr: HRESULT;
  start: TReference_Time;
begin
  // way ahead of you
  Result:= fPreviewing or (not fGraphBuilt);
  if Result then Exit;
  // stop capture pin
  start:= MAX_TIME;
  if ACap<>nil then
    Builder.ControlStream(@PIN_CATEGORY_CAPTURE, @MEDIATYPE_Audio, ACap, @start, nil, 0, 0);

  hr:= Builder.ControlStream(@PIN_CATEGORY_CAPTURE, @MEDIATYPE_Interleaved, VCap, @start, nil, 0, 0);
  if Failed(hr) then
    hr:= Builder.ControlStream(@PIN_CATEGORY_CAPTURE, @MEDIATYPE_Video, VCap, @start, nil, 0, 0);
  if Failed(hr) then Exit; // can't control the stream

  // run the graph
  hr:= Graph.QueryInterface(IID_IMediaControl, MC);
  if Succeeded(hr) then begin
    hr:= MC.Run;
    if Failed(hr) then MC.Stop;  // stop parts that ran
    MC:= nil;
  end;
  if Failed(hr) then begin
    ErrMsg(rsCantRunPreviewGraph, hr);
    Exit;
  end;
  if Assigned(FOnStartPreview) then FOnStartPreview(Self);
  fPreviewing:= true;
  Result:= true;
end;

function TVideoCapture.StopPreview: boolean;
var
  MC: IMediaControl;
  hr: HRESULT;
  start: TReference_Time;
begin
  Result:= false;
  // way ahead of you
  if not fPreviewing then Exit;
  // stop the graph
  MC:= nil;
  if Graph <> nil then begin
    // somehow Preview category won't work ? ... this way we only start the preview .. really :)
    start:= MAX_TIME;
    if ACap<>nil then
      Builder.ControlStream( @PIN_CATEGORY_CAPTURE, @MEDIATYPE_Audio, ACap, @start, nil, 0, 0);

    hr:= Builder.ControlStream(@PIN_CATEGORY_CAPTURE, @MEDIATYPE_Interleaved, VCap, @start, nil, 0, 0);
    if Failed(hr) then
      hr:= Builder.ControlStream(@PIN_CATEGORY_CAPTURE, @MEDIATYPE_Video, VCap, @start, nil, 0, 0);
    if Failed(hr) then Exit; // can't control the stream

    hr:= Graph.QueryInterface(IID_IMediaControl, MC);
    if Succeeded(hr) then begin
      hr:= MC.Stop;
      MC:= nil;
    end;
    if Failed(hr) then begin
      ErrMsg(rsCantStopPreviewGraph, hr);
      Exit;
    end;
  end;
  Invalidate; // !!! get rid of garbage
  if Assigned(FOnStopPreview) then FOnStopPreview(Self);
  fPreviewing:= false;
  Result:= true;
end;

// capture AVI
function TVideoCapture.StartCapture(const Dialog: boolean): boolean;
var
  fHasStreamControl: boolean;
  hr: HRESULT;
  start, stop: TREFERENCE_TIME;
  MC: IMediaControl;
begin
  // way ahead of you
  Result:= false;
  if not FWantCapture then Exit;

  Result:= fCapturing;
  if Result then Exit;
  FCaptureTimer.Enabled:= false;

  // or we'll crash
  Result:= false;
  if not BuildGraph then Exit;

  // This amount will be subtracted from the number of dropped and not
  // dropped frames reported by the filter.  Since we might be having the
  // filter running while the pin is turned off, we don't want any of the
  // frame statistics from the time the pin is off interfering with the
  // statistics we gather while the pin is on
  FDroppedFrames:= 0; FNotDropped:= 0;

  start:= MAX_TIME; stop:= MAX_TIME;

  if ACap<>nil then
    Builder.ControlStream(@PIN_CATEGORY_CAPTURE, @MEDIATYPE_Audio, ACap, @start, nil, 0, 0);
  // don't capture quite yet...
  hr:= Builder.ControlStream(@PIN_CATEGORY_CAPTURE, @MEDIATYPE_Interleaved, VCap, @start, nil, 0, 0);
  if Failed(hr) then
    hr:= Builder.ControlStream(@PIN_CATEGORY_CAPTURE, @MEDIATYPE_Video, VCap, @start, nil, 0, 0);
  // Do we have the ability to control capture and preview separately?
  fHasStreamControl:= SUCCEEDED(hr); // can't control the stream

  // prepare to run the graph
  MC:= nil;
  hr:= Graph.QueryInterface(IID_IMediaControl, MC);
  if (FAILED(hr)) then begin
    ErrMsg(rsCantGetMediaControl, hr);
    Result:= false;
    Exit;
  end;

  try
    try
      // If we were able to keep capture off, then we can
      // run the graph now for frame accurate start later yet still showing a
      // preview. Otherwise, we can't run the graph yet without capture
      // starting too, so we'll pause it so the latency between when they
      // press a key and when capture begins is still small (but they won't have
      // a preview while they wait to press a key)
      if (fHasStreamControl) then hr:= MC.Run()
                             else hr:= MC.Pause();

      if (FAILED(hr)) then // stop parts that started
        ErrMsgException(rsCantRunGraph, hr);

      // press a key to start capture
      if (Dialog and (MessageDlg(rsStartCapture, mtConfirmation, [mbYes, mbCancel], 0) <> mrYes)) then begin
        MC.Stop(); // !!! tnx to Gabriel Corneanu
        if (fWantPreview) then begin
          BuildGraph();
          StartPreview();
        end;
        Abort();
      end;

      // Start capture NOW!
      if (fHasStreamControl) then begin
          // we may not have this yet
          if (DroppedFrames=nil) then begin
            hr:= Builder.FindInterface( @PIN_CATEGORY_CAPTURE, @MEDIATYPE_Interleaved,
                                        VCap, IID_IAMDroppedFrames, DroppedFrames);
            if (hr <> NOERROR) then
              Builder.FindInterface( @PIN_CATEGORY_CAPTURE, @MEDIATYPE_Video,
                                     VCap, IID_IAMDroppedFrames, DroppedFrames);
          end;
          // turn the capture pin on now!
          if ACap<>nil then
            Builder.ControlStream(@PIN_CATEGORY_CAPTURE, @MEDIATYPE_Audio, ACap,  nil, @stop, 0, 0);

          hr:= Builder.ControlStream(@PIN_CATEGORY_CAPTURE, @MEDIATYPE_Interleaved, VCap, nil, @stop, 0, 0);
          if Failed(hr) then
            Builder.ControlStream(@PIN_CATEGORY_CAPTURE, @MEDIATYPE_Video, VCap, nil, @stop, 0, 0);
          // make note of the current dropped frame counts
          if (DroppedFrames<>nil) then begin
            DroppedFrames.GetNumDropped(FDroppedBase);
            DroppedFrames.GetNumNotDropped(FNotDroppedBase);
          end;
      end
      else begin
        hr:= MC.Run();
        if (FAILED(hr)) then // stop parts that started
          ErrMsgException(rsCantRunGraph, hr);
      end;

  // when did we start capture?
      FCapStartTime:= timeGetTime();
      FCaptureTimer.Enabled:= true;
      fCapturing:= true;
      Result:= true;
      if Assigned(FOnStartCapture) then FOnStartCapture(Self);
    except
      MC.Stop();
      Result:= false;
    end
  finally
    MC:= nil;
  end;

end;

function TVideoCapture.StopCapture: boolean;
var
  MC: IMediaControl;
  hr: HRESULT;
begin
  // way ahead of you
  Result:= false;
  if (not fCapturing) then Exit;

  FCaptureTimer.Enabled:= false;

  // stop the graph
  MC:= nil;
  hr:= Graph.QueryInterface(IID_IMediaControl, MC);
  if (SUCCEEDED(hr)) then begin
    hr:= MC.Stop();
    MC:= nil;
  end;
  if (FAILED(hr)) then begin
    ErrMsg(rsCantStopGraph, hr);
    Exit;
  end;

  // when the graph was stopped
  FCapStopTime:= timeGetTime();
  FCapturing:= false;
  fPreviewing:= false;

  if UseTempFile and SaveCaptureFile(CaptureFileName) then begin
    DeleteFile(TempCaptureFileName);
  end;
  // one last time for the final count and all the stats
  try
    if Assigned(FOnCaptureProgress) then FOnCaptureProgress(Self);
  except
  end;
  // !!! get rid of garbage
  Invalidate;

  if Assigned(FOnStopCapture) then FOnStopCapture(Self);

  if fWantPreview then StartPreview();

  Result:= true;
end;

procedure TVideoCapture.CaptureProgress(Sender: TObject);
begin
  FCapTime:= timeGetTime() - FCapStartTime;
  UpdateStatus;
  if Capturing then
    if (UseTimeLimit) and (FCapTime div 1000 >= DWORD(TimeLimit)) then begin
      StopCapture();
      if (FWantPreview) then begin
//        BuildGraph();
        StartPreview();
      end;
    end
    else
      if Assigned(FOnCaptureProgress) then FOnCaptureProgress(Self);
end;

procedure TVideoCapture.UpdateStatus;
begin
  // this filter can't tell us dropped frame info.
  if (DroppedFrames<>nil) and FCapturing then begin
    if (DroppedFrames.GetNumDropped(FDroppedFrames)=S_OK) and
       (DroppedFrames.GetNumNotDropped(FNotDropped)=S_OK) then
    begin
      FDroppedFrames:= FDroppedFrames - FDroppedBase;
      FNotDropped:= FNotDropped - FNotDroppedBase;
    end;
{
    // we want all possible stats, including capture time and actual acheived
    // frame rate and data rate (as opposed to what we tried to get).  These
    // numbers are an indication that though we dropped frames just now, if we
    // chose a data rate and frame rate equal to the numbers I'm about to
    // print, we probably wouldn't drop any frames.
    // average size of frame captured
    if (DroppedFrames.GetAverageFrameSize(&lAvgFrameSize) = S_OK) then begin
    end;
}
  end;
end;

procedure TVideoCapture.SetFrameRate(const Value: double);
const
  DefaultFrameRate = 25;
begin
  if (Value < 1) or (Value > 50) then
    FFrameRate:= DefaultFrameRate
  else
    FFrameRate:= Value;
end;

procedure TVideoCapture.SetTempCaptureFileName(const Value: string);
begin
  FTempCaptureFileName:= Value;
  if (Sink<>nil) then begin
    Sink.SetFileName(PWCHAR(WideString(FTempCaptureFileName)), TAM_Media_Type(nil^));
  end;
end;

function TVideoCapture.AllocCaptureFile(const SizeMb: integer): boolean;
begin
  Result:= MakeBuilder and (Builder.AllocCapFile(PWCHAR(FCaptureFile),
                            SizeMb*1024*1024) = NOERROR);
end;

function TVideoCapture.SaveCaptureFile(const FileName: string): boolean;
var
  tempBuilder: ICaptureGraphBuilder2;
begin
  Result:= FCaptureFile <> '';
  if Result then
    try
      // we need our own graph builder because the main one might not exist
      Result:= (CoCreateInstance( CLSID_CaptureGraphBuilder2, nil, CLSCTX_INPROC,
                                  IID_ICaptureGraphBuilder2,  tempBuilder ) = NOERROR);
      if (Result) then
        try
          // allow the user to press ESC to abort... don't ask for progress
          Result:= tempBuilder.CopyCaptureFile( PWCHAR(FCaptureFile),
                                                PWCHAR(WideString(FileName)),
                                                -1, nil) = S_OK;
        finally
          tempBuilder:= nil;
        end;
    except
      Result:= false;
    end;
end;

procedure TVideoCapture.SetCaptureFileName(const Value: string);
begin
  FCaptureFileName:= Value;
end;

procedure TVideoCapture.GetAvailableDialogs;
var
  Spec: ISpecifyPropertyPages;
  StreamConf: IAMStreamConfig;
  auuid: CAUUID;
  hr: HRESULT;
  TVTuner: IAMTVTuner;
  TVAudio: IAMTVAudio;
begin
  FDialogs:= [];
  if CaptureDialogs <> nil then begin
    // If this device supports the old legacy UI dialogs, offer them
    if CaptureDialogs.HasDialog(VfwCaptureDialog_Format) = S_OK then begin
      FDialogs:= FDialogs + [cdVFORMAT];
    end;
    if CaptureDialogs.HasDialog(VfwCaptureDialog_Source) = S_OK then begin
      FDialogs:= FDialogs + [cdVSOURCE];
    end;
    if CaptureDialogs.HasDialog(VfwCaptureDialog_Display) = S_OK then begin
      FDialogs:= FDialogs + [cdVDISPLAY];
    end;
  end;

  // Also check the audio capture filter at this point, since even non wdm devices
  // may support an IAMAudioInputMixer property page (we'll also get any wdm filter
  // properties here as well). We'll get any audio capture pin property pages just
  // a bit later.
  if (ACap<>nil) and (ACap.QueryInterface(IID_ISpecifyPropertyPages, Spec)=S_OK) then begin
    if (Spec.GetPages(auuid)=S_OK) and (auuid.cElems>0) then begin
      FDialogs:= FDialogs + [cdACAPTURE];
      CoTaskMemFree(auuid.pElems);
    end;
    Spec:= nil;
  end;

  if (VComp<>nil) and (VComp.QueryInterface(IID_ISpecifyPropertyPages, Spec)=S_OK) then begin
    if (Spec.GetPages(auuid)=S_OK) and (auuid.cElems>0) then begin
      FDialogs:= FDialogs + [cdVCOMPRESSION];
      CoTaskMemFree(auuid.pElems);
    end;
    Spec:= nil;
  end;

  // don't bother looking for new property pages if the old ones are supported
  // or if we don't have a capture filter
  if (VCap=nil) or (cdVFORMAT in FDialogs) then Exit;

  // New WDM devices support new UI and new interfaces.
  // Your app can use some default property
  // pages for UI if you'd like (like we do here) or if you don't like our
  // dialog boxes, feel free to make your own and programmatically set
  // the capture options through interfaces like IAMCrossbar, IAMCameraControl
  // etc.
  // There are 9 objects that might support property pages.  Let's go through
  // them.
  // 1. the video capture filter itself

  if (VCap.QueryInterface(IID_ISpecifyPropertyPages, Spec)=S_OK) then begin
    if (Spec.GetPages(auuid)=S_OK) and (auuid.cElems>0) then begin
      FDialogs:= FDialogs + [cdVCAPTURE];
      CoTaskMemFree(auuid.pElems);
    end;
    Spec := nil;
  end;

  // 2.  The video capture capture pin
  hr:= Builder.FindInterface(@PIN_CATEGORY_CAPTURE, @MEDIATYPE_Interleaved, VCap, IID_IAMStreamConfig, StreamConf);
  if hr<>S_OK then
    hr:= Builder.FindInterface(@PIN_CATEGORY_CAPTURE, @MEDIATYPE_Video, VCap, IID_IAMStreamConfig, StreamConf);
  if hr=S_OK then begin
    if StreamConf.QueryInterface(IID_ISpecifyPropertyPages, Spec)=S_OK then begin
      if (Spec.GetPages(auuid)=S_OK) and (auuid.cElems>0) then begin
        FDialogs:= FDialogs + [cdVCAPTURE_PIN];
        CoTaskMemFree(auuid.pElems);
      end;
      Spec:= nil;
    end;
    StreamConf:= nil;
  end;

  // 3.  The video capture preview pin.

  // This basically sets the format being previewed.  Typically, you
  // want to capture and preview using the SAME format, instead of having to
  // enter the same value in 2 dialog boxes.  For a discussion on this, see
  // the comment above the MakePreviewGraph function.
  hr:= Builder.FindInterface(@PIN_CATEGORY_PREVIEW, @MEDIATYPE_Interleaved, VCap, IID_IAMStreamConfig, StreamConf);
  if hr<>S_OK then
    hr:= Builder.FindInterface(@PIN_CATEGORY_PREVIEW, @MEDIATYPE_Video, VCap, IID_IAMStreamConfig, StreamConf);
  if hr = S_OK then begin
    if StreamConf.QueryInterface(IID_ISpecifyPropertyPages, Spec) = S_OK then begin
      if (Spec.GetPages(auuid)=S_OK) and (auuid.cElems>0) then begin
        FDialogs:= FDialogs + [cdVPREVIEW_PIN];
        CoTaskMemFree(auuid.pElems);
      end;
      Spec:= nil;
    end;
    StreamConf:= nil;
  end;

  // 4 & 5.  The video crossbar, and a possible second crossbar

  // 6.  The TVTuner
  hr:= Builder.FindInterface(@PIN_CATEGORY_CAPTURE, @MEDIATYPE_Interleaved, VCap, IID_IAMTVTuner, TVTuner);
  if hr <> S_OK then
    hr:= Builder.FindInterface(@PIN_CATEGORY_CAPTURE, @MEDIATYPE_Video, VCap, IID_IAMTVTuner, TVTuner);
  if hr=S_OK then begin
    if TVTuner.QueryInterface(IID_ISpecifyPropertyPages, Spec)=S_OK then begin
      if (Spec.GetPages(auuid)=S_OK) and (auuid.cElems>0) then begin
        FDialogs:= FDialogs + [cdTVTUNER];
        CoTaskMemFree(auuid.pElems);
      end;
      Spec:= nil;
    end;
    TVTuner:= nil;
  end;

  // no audio capture, we're done
  if ACap = nil then Exit;

  if (AComp<>nil) and (AComp.QueryInterface(IID_ISpecifyPropertyPages, Spec)=S_OK) then begin
    if (Spec.GetPages(auuid)=S_OK) and (auuid.cElems>0) then begin
      FDialogs:= FDialogs + [cdACOMPRESSION];
      CoTaskMemFree(auuid.pElems);
    end;
    Spec:= nil;
  end;

  // 7.  The Audio capture filter itself... Thanks anyway, but we got these already

  // 8.  The Audio capture pin
  if Builder.FindInterface(@PIN_CATEGORY_CAPTURE, @MEDIATYPE_Audio, ACap, IID_IAMStreamConfig, StreamConf)=S_OK then begin
    if StreamConf.QueryInterface(IID_ISpecifyPropertyPages, Spec)=S_OK then begin
      if (Spec.GetPages(auuid)=S_OK) and (auuid.cElems>0) then begin
        FDialogs:= FDialogs + [cdACAPTURE_PIN];
        CoTaskMemFree(auuid.pElems);
      end;
      Spec:= nil;
    end;
    StreamConf:= nil;
  end;

  // 9.  The TV Audio filter
  if Builder.FindInterface(@PIN_CATEGORY_CAPTURE, @MEDIATYPE_Audio, ACap, IID_IAMTVAudio, TVAudio)=S_OK then begin
    if TVAudio.QueryInterface(IID_ISpecifyPropertyPages, Spec)=S_OK then begin
      if (Spec.GetPages(auuid)=S_OK) and (auuid.cElems>0) then begin
        FDialogs:= FDialogs + [cdTVAUDIO];
        CoTaskMemFree(auuid.pElems);
      end;
      Spec:= nil;
    end;
    TVAudio:= nil;
  end;
end;

function TVideoCapture.ShowDialog(const DialogType: TCaptureDialog): boolean;
var
  Spec: ISpecifyPropertyPages;
  auuid: CAUUID;
  hr: HRESULT;
  StreamConf: IAMStreamConfig;
  TVTuner: IAMTVTuner;
begin
  Result:= false;
  if not (DialogType in FDialogs) or FCapturing then Exit;
  // they want the VfW format dialog
  case DialogType of
    cdVFORMAT: begin
        // this dialog will not work while previewing
        // Oh uh!  Sometimes bringing up the FORMAT dialog can result
        // in changing to a capture format that the current graph
        // can't handle.  It looks like that has happened and we'll
        // have to rebuild the graph.
        if Cardinal(CaptureDialogs.ShowDialog(VfwCaptureDialog_Format, Handle))=VFW_E_CANNOT_CONNECT then
          TearDownGraph;    // now we need to rebuild
        if fWantPreview then begin
          BuildGraph;
          StartPreview;
        end;
      end;
    cdVSOURCE: begin
        // this dialog will not work while previewing
        if fWantPreview then StopPreview;
        CaptureDialogs.ShowDialog(VfwCaptureDialog_Source, Handle );
        if fWantPreview then StartPreview;
      end;
    cdVDISPLAY: begin
        // this dialog will not work while previewing
        if fWantPreview then StopPreview;
        CaptureDialogs.ShowDialog(VfwCaptureDialog_Display, Handle);
        if fWantPreview then StartPreview;
      end;
    // now the code for the new dialogs
    cdVCAPTURE: begin
        if VCap.QueryInterface(IID_ISpecifyPropertyPages, Spec)=S_OK then begin
          Spec.GetPages(auuid);
          OleCreatePropertyFrame(Handle, 30, 30, nil, 1, @VCap, auuid.cElems, auuid.pElems, 0, 0, nil);
          CoTaskMemFree(auuid.pElems);
          Spec:= nil;
        end;
      end;
    cdVCAPTURE_PIN: begin
        // You can change this pin's output format in these dialogs.
        // If the capture pin is already connected to somebody who's
        // fussy about the connection type, that may prevent using
        // this dialog(!) because the filter it's connected to might not
        // allow reconnecting to a new format. (EG: you switch from RGB
        // to some compressed type, and need to pull in a decoder)
        // I need to tear down the graph downstream of the
        // capture filter before bringing up these dialogs.
        // In any case, the graph must be STOPPED when calling them.
        // The capture pin that we are trying to set the format on is connected if
        // one of these variable is set to TRUE. The pin should be disconnected for
        // the dialog to work properly.
        TearDownGraph;    // graph could prevent dialog working
        hr:= Builder.FindInterface(@PIN_CATEGORY_CAPTURE, @MEDIATYPE_Interleaved, VCap, IID_IAMStreamConfig, StreamConf);
        if hr<>S_OK then
          Builder.FindInterface(@PIN_CATEGORY_CAPTURE, @MEDIATYPE_Video, VCap, IID_IAMStreamConfig, StreamConf);
        if StreamConf.QueryInterface(IID_ISpecifyPropertyPages, Spec)=S_OK then begin
          Spec.GetPages(auuid);
          OleCreatePropertyFrame(Handle, 30, 30, nil, 1, @StreamConf, auuid.cElems, auuid.pElems, 0, 0, nil);
          CoTaskMemFree(auuid.pElems);
          Spec:= nil;
        end;
        StreamConf:= nil;
        if fWantPreview then begin
          BuildGraph;
          StartPreview;
        end;
      end;
    cdVPREVIEW_PIN: begin
        // this dialog may not work if the preview pin is connected
        // already, because the downstream filter may reject a format
        // change, so we better kill the graph. (EG: We switch from
        // capturing RGB to some compressed fmt, and need to pull in
        // a decompressor)
        TearDownGraph;
        // This dialog changes the preview format, so it might affect
        // the format being drawn.  Our app's window size is taken
        // from the size of the capture pin's video, not the preview
        // pin, so changing that here won't have any effect. All in all,
        // this probably won't be a terribly useful dialog in this app.
        hr:= Builder.FindInterface(@PIN_CATEGORY_PREVIEW, @MEDIATYPE_Interleaved, VCap, IID_IAMStreamConfig, StreamConf);
        if hr<>S_OK then
          Builder.FindInterface(@PIN_CATEGORY_PREVIEW, @MEDIATYPE_Video, VCap, IID_IAMStreamConfig, StreamConf);
        if StreamConf.QueryInterface(IID_ISpecifyPropertyPages, Spec)=S_OK then begin
          Spec.GetPages(auuid);
          OleCreatePropertyFrame(Handle, 30, 30, nil, 1, @StreamConf, auuid.cElems, auuid.pElems, 0, 0, nil);
          CoTaskMemFree(auuid.pElems);
          Spec:= nil;
        end;
        StreamConf:= nil;
        BuildGraph;
        if fWantPreview then StartPreview;
      end;
    cdTVTUNER: begin
        hr:= Builder.FindInterface(@PIN_CATEGORY_CAPTURE, @MEDIATYPE_Interleaved, VCap, IID_IAMTVTuner, TVTuner);
        if hr<>S_OK then
          Builder.FindInterface(@PIN_CATEGORY_CAPTURE, @MEDIATYPE_Video, VCap, IID_IAMTVTuner, TVTuner);
        if TVTuner.QueryInterface(IID_ISpecifyPropertyPages, Spec)=S_OK then begin
          Spec.GetPages(auuid);
          OleCreatePropertyFrame(Handle, 30, 30, nil, 1, @TVTuner, auuid.cElems, auuid.pElems, 0, 0, nil);
          CoTaskMemFree(auuid.pElems);
          Spec:= nil;
        end;
        TVTuner:= nil;
      end;
    cdACAPTURE: begin
        if ACap.QueryInterface(IID_ISpecifyPropertyPages, Spec)=S_OK then begin
          Spec.GetPages(auuid);
          OleCreatePropertyFrame(Handle, 30, 30, nil, 1, @ACap, auuid.cElems, auuid.pElems, 0, 0, nil);
          CoTaskMemFree(auuid.pElems);
          Spec:= nil;
        end;
      end;
    cdACAPTURE_PIN: begin
        // this dialog will not work while previewing - it might change
        // the output format!
        if fWantPreview then StopPreview;
        Builder.FindInterface(@PIN_CATEGORY_CAPTURE, @MEDIATYPE_Audio, ACap, IID_IAMStreamConfig, StreamConf);
        if StreamConf.QueryInterface(IID_ISpecifyPropertyPages, Spec)=S_OK then begin
          Spec.GetPages(auuid);
          OleCreatePropertyFrame(Handle, 30, 30, nil, 1, @StreamConf, auuid.cElems, auuid.pElems, 0, 0, nil);
          CoTaskMemFree(auuid.pElems);
          Spec:= nil;
        end;
        StreamConf:= nil;
        if fWantPreview then StartPreview;
      end;
    cdVCOMPRESSION: begin
        if VComp.QueryInterface(IID_ISpecifyPropertyPages, Spec)=S_OK then begin
          Spec.GetPages(auuid);
          OleCreatePropertyFrame(Handle, 30, 30, nil, 1, @VComp, auuid.cElems, auuid.pElems, 0, 0, nil);
          CoTaskMemFree(auuid.pElems);
          Spec:= nil;
        end;
      end;
    cdACOMPRESSION: begin
        if AComp.QueryInterface(IID_ISpecifyPropertyPages, Spec)=S_OK then begin
          Spec.GetPages(auuid);
          OleCreatePropertyFrame(Handle, 30, 30, nil, 1, @AComp, auuid.cElems, auuid.pElems, 0, 0, nil);
          CoTaskMemFree(auuid.pElems);
          Spec:= nil;
        end;
      end;
  end;
  Result:= true;
end;

initialization
  VideoDevicesList:= TObjectList.Create;
  AudioDevicesList:= TObjectList.Create;
  VideoCompressorsList:= TObjectList.Create;
  AudioCompressorsList:= TObjectList.Create;

finalization
  VideoDevicesList.Free;
  AudioDevicesList.Free;
  VideoCompressorsList.Free;
  AudioCompressorsList.Free;

end.
